﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Linq;
using System.Text;
using FDAL.Database;

namespace FDAL.Gen
{
    public class TMethodGenerator
    {
        private const string USING_STRING =
@"using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Linq;
using System.Text;
using KDAL;";

        private TCodeWriter writer;
        private TDatabaseExplorer explorer;

        public TMethodGenerator(SqlConnection conn)
        {
            this.writer = new TCodeWriter();
            this.explorer = new TDatabaseExplorer(conn);
        }

        // gets the parameter definition string in method header
        private string GetParamDefString(SqlParameter p)
        {
            Type ptype = TypeMapper.GetCLRType(p.SqlDbType);
            string ptypeDef = ptype.Name;
            
            // handle nullable value types
            if (p.Direction == ParameterDirection.Input && ptype.IsValueType)
            {
                ptypeDef += "?";
            }
            string pname = TNameConvert.GetPascalString(p.ParameterName).Replace("@", "");
            string piden = null;
            switch (p.Direction)
            {
                case ParameterDirection.InputOutput:
                    piden = "ref ";
                    break;
                case ParameterDirection.Output:
                    piden = "out ";
                    break;
                case ParameterDirection.Input:
                    piden = "";
                    break;
            }
            string defString = string.Format("{0}{1} {2}", piden, ptypeDef, pname);  
            return defString;
        }

        
        // gets the paramater declaration string in method body
        private string GetParamDecString(SqlParameter p)
        {
            string pName = TNameConvert.GetPascalString(p.ParameterName).Replace("@", "");
            string s1 = string.Format("SqlParameter p{0} = new SqlParameter();", pName);
            string s2 = string.Format("p{0}.ParameterName = \"{1}\";", pName, p.ParameterName);
            string s3 = string.Format("p{0}.SqlDbType = SqlDbType.{1};", pName, p.SqlDbType);
            string s4 = string.Format("p{0}.Size = {1};", pName, p.Size);
            string s5 = string.Format("p{0}.Value = ({0} == null)? (object)DBNull.Value : {0};", pName);
            string s6 = string.Format("p{0}.Direction = ParameterDirection.{1};", pName, p.Direction);
            string s7 = string.Format("parameters.Add(p{0});", pName);
            StringBuilder sb = new StringBuilder();
            sb.AppendLine(s1);
            sb.AppendLine(s2);
            sb.AppendLine(s3);
            sb.AppendLine(s4);
            sb.AppendLine(s5);
            sb.AppendLine(s6);
            sb.AppendLine(s7);
            return sb.ToString();
        }

        public string GenerateMethod(TStoredProcedure storedProc)
        {
            SqlParameterCollection parameters = explorer.GetStoredProcParameters(storedProc);
            List<string> paramDefs = new List<string>();
            foreach(SqlParameter p in parameters)
            {
                if (p.Direction != ParameterDirection.ReturnValue)
                {
                    paramDefs.Add(GetParamDefString(p));
                }
            }

            paramDefs.Add("params Type[] resultTypes");
            
            string paramDef = TStringUtils.GenerateCommaSeparatedString(paramDefs, null);
            string returnTypeDef = "TStoredProcResult";
            string methodName = TNameConvert.GetPascalString(storedProc.Name);
            string methodHeaderDef = string.Format("public {0} {1}({2})", returnTypeDef, methodName, paramDef);

            writer.WriteLine(methodHeaderDef);
            writer.WriteLine("{");
            writer.PushIdent();

            if (paramDefs.Count > 1)
            {
                writer.WriteLine("List<SqlParameter> parameters = new List<SqlParameter>();");
                foreach (SqlParameter p in parameters)
                {
                    if (p.Direction != ParameterDirection.ReturnValue)
                    {
                        string s = GetParamDecString(p);
                        writer.WriteLine(s);
                    }
                }

                string se = string.Format("TStoredProcResult result = ExecProc(\"{0}\", parameters, resultTypes);", storedProc.Name);
                writer.WriteLine(se);
            }
            else
            {
                string se = string.Format("TStoredProcResult result = ExecProc(\"{0}\", null, resultTypes);", storedProc.Name);
                writer.WriteLine(se);
            }

            int index = 0;
            foreach (SqlParameter p in parameters)
            {
                if (p.Direction == ParameterDirection.InputOutput || p.Direction == ParameterDirection.Output)
                {
                    string pName = TNameConvert.GetPascalString(p.ParameterName).Replace("@", "");
                    Type ptype = TypeMapper.GetCLRType(p.SqlDbType);
                    string ptypeDef = ptype.Name;
                    writer.WriteLineFormat("{0} = ({1})result.OutPramaters.ElementAt<SqlParameter>({2}).Value;", pName, ptypeDef, index);
                    index++;
                }
            }

            writer.WriteLine("return result;");
            
            writer.PopIdent();
            writer.WriteLine("}");
            return writer.Code;
        }

        public string GenerateClass(string namespaceName, string className)
        {
            writer.WriteLine("// KDAL Generated Code");
            writer.WriteLineFormat(string.Format("//Date: {0}", DateTime.Now));
            writer.WriteLine();
            writer.WriteLine(USING_STRING);
            writer.WriteLine("");
            writer.WriteLineFormat("namespace {0}", namespaceName);
            writer.WriteLine("{");
            writer.PushIdent();
            writer.WriteLineFormat("public class {0} : TContext", className);
            writer.WriteLine("{");
            writer.PushIdent();
            writer.WriteLineFormat("public {0}(SqlConnection connection, SqlTransaction transaction) : base(connection, transaction) {{ }}", className);
            writer.WriteLineFormat("public {0}(SqlConnection connection) : base(connection, null) {{ }}", className);

            writer.WriteLine("");
            List<TStoredProcedure> procs = explorer.GetStoredProcedures();
            foreach (TStoredProcedure proc in procs)
            {
                GenerateMethod(proc);
                writer.WriteLine("");
            }



            writer.PopIdent();
            writer.WriteLine("}");
            writer.PopIdent();
            writer.WriteLine("}");

            return writer.Code;
        }
    }
}
