﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Data.SqlClient;
using FDAL.Database;

namespace FDAL.Gen
{
    public class TStoredProcGenerator
    {
        private TDatabaseExplorer explorer;
        private string database;

        public TStoredProcGenerator(SqlConnection conn, string database)
        {
            this.explorer = new TDatabaseExplorer(conn);
            this.database = database;
        }

        public string GetColumnTypeString(TColumn colInfo)
        {
            string dataType = colInfo.DataType.ToLower();
            string result = "";
            string[] charTypes = { "char", "varchar", "nchar", "nvarchar" };
            string[] binaryTypes = { "binary", "varbinary" };
            string[] numericType = { "decimal", "numeric" };
            string[] dateTypes = { "datetime2", "time", "datetimeoffset"};
            
            if (charTypes.Contains<string>(dataType) || binaryTypes.Contains<string>(dataType))
            {
                string len = colInfo.CharMaxLength == null ? "" : colInfo.CharMaxLength == -1 ? "(MAX)" : string.Format("({0})", colInfo.CharMaxLength);
                result = colInfo.DataType + len;
            }
            else if (numericType.Contains<string>(dataType))
            {
                result = string.Format("{0}({1}, {2})", colInfo.DataType, colInfo.NumericPrecision, colInfo.NumericScale);
            }
            else if (dateTypes.Contains<string>(dataType))
            {
                result = string.Format("{0}({1})", colInfo.DataType, colInfo.DateTimePrecision);
            }
            else
            {
                result = dataType;
            }
            return result;
        }

        private string GetParametersString(List<TColumn> cols)
        {
            TCodeWriter writer = new TCodeWriter();
            int maxLength = cols.Max<TColumn>(ci => ci.Name.Length);
            // create a format string according to maximum field name
            string f = string.Format("@{{0,-{0}}}{{1,18}}", maxLength);

            for (int i = 0; i < cols.Count; i++)
            {
                TColumn inf = cols[i];
                writer.WriteFormat(f, inf.Name, GetColumnTypeString(inf));
                if (i < cols.Count - 1)
                    writer.WriteNoIdent(",");
                writer.WriteLine();
            }
            return writer.Code;
        }

        private string GetNameFromColumns(List<TColumn> cols)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < cols.Count; i++)
            {
                TColumn col = cols[i];
                sb.Append(col.Name);
                if (i < cols.Count - 1)
                {
                    sb.Append("_AND_");
                }
            }
            return sb.ToString();
        }

        private string GetWhereFromColumns(List<TColumn> cols)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < cols.Count; i++)
            {
                TColumn col = cols[i];
                sb.AppendFormat("({0} = @{0})", col.Name);
                if (i < cols.Count - 1)
                {
                    
                    sb.Append(" AND ");
                }
            }
            return sb.ToString();
        }

        //=============================== INSERT ===========================
        public string GenerateDropIfExists(string spName)
        {
            TCodeWriter writer = new TCodeWriter();
            writer.WriteLineFormat("IF EXISTS (SELECT * FROM sys.objects WHERE object_id = OBJECT_ID(N'{0}') AND type in (N'P', N'PC'))", spName);
            writer.PushIdent();
            writer.WriteLineFormat("DROP PROCEDURE {0};", spName);
            writer.PopIdent();
            writer.WriteLine("GO");
            return writer.Code;
        }
        
        public string GenerateInsert(TTable table)
        {
            TCodeWriter writer = new TCodeWriter();
            string spName = string.Format("INSERT_{0}", table.Name);
            writer.WriteLine(GenerateDropIfExists(spName));
            writer.WriteLineFormat("CREATE PROCEDURE {0}", spName);
            List<TColumn> colInfos = explorer.GetColumns(table).Where<TColumn>(ci => !(ci.IsComputed || ci.IsIdentity)).ToList<TColumn>();
            int maxLength = colInfos.Max<TColumn>(ci => ci.Name.Length);
            // create a format string according to maximum field name
            string f = string.Format("@{{0,-{0}}}{{1,18}}", maxLength);

            for (int i = 0; i < colInfos.Count; i++)
            {
                TColumn inf = colInfos[i];
                writer.WriteFormat(f, inf.Name, GetColumnTypeString(inf));
                if (i < colInfos.Count - 1)
                    writer.WriteNoIdent(",");
                writer.WriteLine();
            }
            writer.WriteLine("AS");
            writer.WriteLine("BEGIN");
            writer.PushIdent();
            writer.WriteLineFormat("INSERT INTO {0}", table.SafeName);
            writer.Write("(");
            for (int i = 0; i < colInfos.Count; i++)
            {
                TColumn inf = colInfos[i];
                writer.WriteFormatNoIdent("[{0}]", inf.Name);
                if (i < colInfos.Count - 1)
                    writer.WriteNoIdent(", ");
            }
            writer.WriteLineNoIdent(")");
            writer.WriteLine("VALUES");
            writer.Write("(");
            for (int i = 0; i < colInfos.Count; i++)
            {
                TColumn inf = colInfos[i];
                writer.WriteNoIdent("@" + inf.Name);
                if (i < colInfos.Count - 1)
                    writer.WriteNoIdent(", ");
            }
            writer.WriteLineNoIdent(")");
            writer.PopIdent();
            writer.WriteLine("END");
            writer.WriteLine("GO");
            return writer.Code;
        }

        //======================= GET ============================
        public string GenerateGet(TTable table, List<TColumn> cols)
        {
            TCodeWriter writer = new TCodeWriter();
            string spName = string.Format("GET_{0}_BY_{1}", table.Name, GetNameFromColumns(cols));
            writer.WriteLine(GenerateDropIfExists(spName));
            writer.WriteLineFormat("CREATE PROCEDURE {0}", spName);
            writer.Write(GetParametersString(cols));
            writer.WriteLine("AS");
            writer.WriteLine("BEGIN");
            writer.PushIdent();
            writer.WriteLineFormat("SELECT * FROM {0}", table.SafeName);
            writer.WriteLine("WHERE ");
            writer.WriteLine(GetWhereFromColumns(cols));
            
            writer.PopIdent();
            writer.WriteLine("END");
            writer.WriteLine("GO");
            writer.WriteLine();
            return writer.Code;
        }

        public string GenerateGetAll(TTable table)
        {
            TCodeWriter writer = new TCodeWriter();
            string spName = string.Format("GET_ALL_{0}S", table.Name);
            writer.WriteLine(GenerateDropIfExists(spName));
            writer.WriteLineFormat("CREATE PROCEDURE {0}", spName);
            writer.WriteLine("AS");
            writer.WriteLine("BEGIN");
            writer.PushIdent();
            writer.WriteLineFormat("SELECT * FROM {0};", table.SafeName);
            writer.PopIdent();
            writer.WriteLine("END");
            writer.WriteLine("GO");
            writer.WriteLine();
            return writer.Code;
        }

        public string GenerateGetByConstraints(TTable table)
        {
            TCodeWriter writer = new TCodeWriter();
            List<TColumn> allColumns = explorer.GetColumns(table);
            List<TConstraint> constraints = explorer.GetConstraints(table);
            foreach (var constraint in constraints)
            {
                if (constraint.Type.Equals("PRIMARY KEY") || constraint.Type.Equals("UNIQUE"))
                {
                    string s = GenerateGet(table, constraint.Columns);
                    writer.Write(s);
                }
            }
            
            return writer.Code;
        }


        //======================= UPDATE ============================
        public string GenerateUpdateBy(TTable table, List<TColumn> cols)
        {
            TCodeWriter writer = new TCodeWriter();
            
            List<TColumn> allColumns = explorer.GetColumns(table);
            List<TColumn> editableCols = allColumns.Where<TColumn>(ci => (!(ci.IsComputed || ci.IsIdentity))).ToList<TColumn>();

            string spName = string.Format("UPDATE_{0}_BY_{1}", table.Name, GetNameFromColumns(cols));
            writer.WriteLine(GenerateDropIfExists(spName));
            writer.WriteLineFormat("CREATE PROCEDURE {0}", spName);
            writer.Write(GetParametersString(allColumns));
            writer.WriteLine("AS");
            writer.WriteLine("BEGIN");
            writer.PushIdent();
            writer.WriteLineFormat("UPDATE {0}", table.SafeName);
            writer.WriteLine("SET");
            writer.PushIdent();
            for (int i = 0; i < editableCols.Count; i++)
            {
                TColumn inf = editableCols[i];
                writer.WriteFormat("[{0}] = @{0}", inf.Name);
                if (i < editableCols.Count - 1)
                    writer.WriteLineNoIdent(", ");
            }
            writer.PopIdent();
            writer.WriteLine();
            writer.WriteLine("WHERE ");
            writer.WriteLine(GetWhereFromColumns(cols));
            writer.PopIdent();
            writer.WriteLine("END");
            writer.WriteLine("GO");
            writer.WriteLine();
            return writer.Code;
        }

        public string GenerateUpdate(TTable table)
        {
            TCodeWriter writer = new TCodeWriter();
            List<TColumn> allColumns = explorer.GetColumns(table);
            
            List<TConstraint> constraints = explorer.GetConstraints(table);
            foreach (var constraint in constraints)
            {
                if (constraint.Type.Equals("PRIMARY KEY") || constraint.Type.Equals("UNIQUE"))
                {
                    string s = GenerateUpdateBy(table, constraint.Columns);
                    writer.Write(s);
                }
            }
            return writer.Code;
        }

        //======================= DELETE ============================
        public string GenerateDeleteBy(TTable table, List<TColumn> cols)
        {
            TCodeWriter writer = new TCodeWriter();
            List<TColumn> allColumns = explorer.GetColumns(table);
            List<TColumn> editableCols = allColumns.Where<TColumn>(ci => (!(ci.IsComputed || ci.IsIdentity))).ToList<TColumn>();
            string spName = string.Format("DELETE_{0}_BY_{1}", table.Name, GetNameFromColumns(cols));
            writer.WriteLine(GenerateDropIfExists(spName));
            writer.WriteLineFormat("CREATE PROCEDURE {0}", spName);
            writer.Write(GetParametersString(cols));
            writer.WriteLine("AS");
            writer.WriteLine("BEGIN");
            writer.PushIdent();
            writer.WriteLineFormat("DELETE FROM {0}", table.SafeName);
            writer.WriteLine("WHERE ");
            writer.WriteLine(GetWhereFromColumns(cols));
            writer.PopIdent();
            writer.WriteLine("END");
            writer.WriteLine("GO");
            writer.WriteLine();
            return writer.Code;
        }

        public string GenerateDelete(TTable table)
        {
            TCodeWriter writer = new TCodeWriter();
            List<TColumn> allColumns = explorer.GetColumns(table);
            List<TConstraint> constraints = explorer.GetConstraints(table);
            foreach (var constraint in constraints)
            {
                if (constraint.Type.Equals("PRIMARY KEY") || constraint.Type.Equals("UNIQUE"))
                {
                    string s = GenerateDeleteBy(table, constraint.Columns);
                    writer.Write(s);
                }
            }
            return writer.Code;
        }

        public string Generate()
        {
            TCodeWriter writer = new TCodeWriter();
            writer.WriteLine("--- KDAL Generated Code");
            writer.WriteLineFormat(string.Format("--- Date: {0}", DateTime.Now));
            writer.WriteLine();
            List<TTable> tables = explorer.GetTables();
            foreach (TTable t in tables)
            {
                writer.WriteLineFormat("----------------- {0} -------------------", t.FullName);
                writer.WriteLineFormat("USE {0};", database);
                writer.WriteLine("GO");
                writer.WriteLine("--- Get ---");
                writer.WriteLine(GenerateGetAll(t));
                writer.WriteLine(GenerateGetByConstraints(t));
                writer.WriteLine("--- Insert ---");
                writer.WriteLine(GenerateInsert(t));
                writer.WriteLine("--- Update ---");
                writer.WriteLine(GenerateUpdate(t));
                writer.WriteLine("--- Delete ---");
                writer.WriteLine(GenerateDelete(t));
            }
            return writer.Code;
        }
    }
}
